@@ -22,8 +22,9 @@ module Codec.Compression.Zlib (
   -- * Simple compression and decompression
   compress,
   decompress,
+  DecompressError(..),
 
-  -- * Extended api with control over compression parameters
+  -- * Extended API with control over compression parameters
   compressWith,
   decompressWith,
 
@@ -37,7 +38,7 @@ module Codec.Compression.Zlib (
     bestSpeed,
     bestCompression,
     compressionLevel,
-  Method(..),
+  Method,
     deflateMethod,
   WindowBits(..),
     defaultWindowBits,
@@ -47,10 +48,12 @@ module Codec.Compression.Zlib (
     minMemoryLevel,
     maxMemoryLevel,
     memoryLevel,
-  CompressionStrategy(..),
+  CompressionStrategy,
     defaultStrategy,
     filteredStrategy,
     huffmanOnlyStrategy,
+    rleStrategy,
+    fixedStrategy,
 
   ) where
 
@@ -60,16 +63,8 @@ import qualified Codec.Compression.Zlib.Internal as Internal
 import Codec.Compression.Zlib.Internal hiding (compress, decompress)
 
 
--- | Decompress a stream of data in the zlib format.
---
--- There are a number of errors that can occur. In each case an exception will
--- be thrown. The possible error conditions are:
---
--- * if the stream does not start with a valid gzip header
---
--- * if the compressed stream is corrupted
---
--- * if the compressed stream ends prematurely
+-- | Decompress a stream of data in the zlib format,
+-- throw 'DecompressError' on failure.
 --
 -- Note that the decompression is performed /lazily/. Errors in the data stream
 -- may not be detected until the end of the stream is demanded (since it is
@@ -81,7 +76,7 @@ decompress :: ByteString -> ByteString
 decompress = decompressWith defaultDecompressParams
 
 
--- | Like 'decompress' but with the ability to specify various decompression
+-- | Like 'Codec.Compression.Zlib.decompress' but with the ability to specify various decompression
 -- parameters. Typical usage:
 --
 -- > decompressWith defaultCompressParams { ... }
@@ -103,7 +98,7 @@ compress :: ByteString -> ByteString
 compress = compressWith defaultCompressParams
 
 
--- | Like 'compress' but with the ability to specify various compression
+-- | Like 'Codec.Compression.Zlib.compress' but with the ability to specify various compression
 -- parameters. Typical usage:
 --
 -- > compressWith defaultCompressParams { ... }
