@@ -1,7 +1,6 @@
-{-# LANGUAGE CPP, RankNTypes, DeriveDataTypeable, BangPatterns #-}
-#if __GLASGOW_HASKELL__ >= 702
+{-# LANGUAGE CPP, RankNTypes, BangPatterns #-}
+{-# LANGUAGE DeriveGeneric #-}
 {-# LANGUAGE Trustworthy #-}
-#endif
 -----------------------------------------------------------------------------
 -- |
 -- Copyright   :  (c) 2006-2015 Duncan Coutts
@@ -45,7 +44,7 @@ module Codec.Compression.Zlib.Internal (
   defaultCompressParams,
   DecompressParams(..),
   defaultDecompressParams,
-  Stream.Format(..),
+  Stream.Format,
     Stream.gzipFormat,
     Stream.zlibFormat,
     Stream.rawFormat,
@@ -56,7 +55,7 @@ module Codec.Compression.Zlib.Internal (
     Stream.bestSpeed,
     Stream.bestCompression,
     Stream.compressionLevel,
-  Stream.Method(..),
+  Stream.Method,
     Stream.deflateMethod,
   Stream.WindowBits(..),
     Stream.defaultWindowBits,
@@ -66,10 +65,12 @@ module Codec.Compression.Zlib.Internal (
     Stream.minMemoryLevel,
     Stream.maxMemoryLevel,
     Stream.memoryLevel,
-  Stream.CompressionStrategy(..),
+  Stream.CompressionStrategy,
     Stream.defaultStrategy,
     Stream.filteredStrategy,
     Stream.huffmanOnlyStrategy,
+    Stream.rleStrategy,
+    Stream.fixedStrategy,
 
   ) where
 
@@ -78,17 +79,16 @@ import Control.Monad (when)
 import Control.Exception (Exception, throw, assert)
 import Control.Monad.ST.Lazy hiding (stToIO)
 import Control.Monad.ST.Strict (stToIO)
-#if __GLASGOW_HASKELL__ >= 702
 import qualified Control.Monad.ST.Unsafe as Unsafe (unsafeIOToST)
-#else
-import qualified Control.Monad.ST.Strict as Unsafe (unsafeIOToST)
-#endif
-import Data.Typeable (Typeable)
+import GHC.Generics (Generic)
+import Data.Bits (toIntegralSized)
 import qualified Data.ByteString.Lazy as L
 import qualified Data.ByteString.Lazy.Internal as L
 import qualified Data.ByteString          as S
 import qualified Data.ByteString.Internal as S
+import Data.Maybe (fromMaybe)
 import Data.Word (Word8)
+import Foreign.C (CUInt)
 import GHC.IO (noDuplicate)
 
 import qualified Codec.Compression.Zlib.Stream as Stream
@@ -113,7 +113,12 @@ data CompressParams = CompressParams {
   compressStrategy    :: !Stream.CompressionStrategy,
   compressBufferSize  :: !Int,
   compressDictionary  :: Maybe S.ByteString
-} deriving Show
+  } deriving
+  ( Eq       -- ^ @since 0.7.0.0
+  , Ord      -- ^ @since 0.7.0.0
+  , Show
+  , Generic  -- ^ @since 0.7.0.0
+  )
 
 -- | The full set of parameters for decompression. The defaults are
 -- 'defaultDecompressParams'.
@@ -139,10 +144,16 @@ data DecompressParams = DecompressParams {
   decompressBufferSize :: !Int,
   decompressDictionary :: Maybe S.ByteString,
   decompressAllMembers :: Bool
-} deriving Show
+  } deriving
+  ( Eq       -- ^ @since 0.7.0.0
+  , Ord      -- ^ @since 0.7.0.0
+  , Show
+  , Generic  -- ^ @since 0.7.0.0
+  )
 
 -- | The default set of parameters for compression. This is typically used with
--- the @compressWith@ function with specific parameters overridden.
+-- 'Codec.Compression.GZip.compressWith' or 'Codec.Compression.Zlib.compressWith'
+-- with specific parameters overridden.
 --
 defaultCompressParams :: CompressParams
 defaultCompressParams = CompressParams {
@@ -151,17 +162,18 @@ defaultCompressParams = CompressParams {
   compressWindowBits  = Stream.defaultWindowBits,
   compressMemoryLevel = Stream.defaultMemoryLevel,
   compressStrategy    = Stream.defaultStrategy,
-  compressBufferSize  = defaultCompressBufferSize,
+  compressBufferSize  = cuint2int defaultCompressBufferSize,
   compressDictionary  = Nothing
 }
 
 -- | The default set of parameters for decompression. This is typically used with
--- the @compressWith@ function with specific parameters overridden.
+-- 'Codec.Compression.GZip.decompressWith' or 'Codec.Compression.Zlib.decompressWith'
+-- with specific parameters overridden.
 --
 defaultDecompressParams :: DecompressParams
 defaultDecompressParams = DecompressParams {
   decompressWindowBits = Stream.defaultWindowBits,
-  decompressBufferSize = defaultDecompressBufferSize,
+  decompressBufferSize = cuint2int defaultDecompressBufferSize,
   decompressDictionary = Nothing,
   decompressAllMembers = True
 }
@@ -169,9 +181,9 @@ defaultDecompressParams = DecompressParams {
 -- | The default chunk sizes for the output of compression and decompression
 -- are 16k and 32k respectively (less a small accounting overhead).
 --
-defaultCompressBufferSize, defaultDecompressBufferSize :: Int
-defaultCompressBufferSize   = 16 * 1024 - L.chunkOverhead
-defaultDecompressBufferSize = 32 * 1024 - L.chunkOverhead
+defaultCompressBufferSize, defaultDecompressBufferSize :: CUInt
+defaultCompressBufferSize   = 16 * 1024 - int2cuint L.chunkOverhead
+defaultDecompressBufferSize = 32 * 1024 - int2cuint L.chunkOverhead
 
 -- | The unfolding of the decompression process, where you provide a sequence
 -- of compressed data chunks as input and receive a sequence of uncompressed
@@ -179,7 +191,7 @@ defaultDecompressBufferSize = 32 * 1024 - L.chunkOverhead
 -- input and provision of output are interleaved.
 --
 -- To indicate the end of the input supply an empty input chunk. Note that
--- for 'gzipFormat' with the default 'decompressAllMembers' @True@ you will
+-- for 'Stream.gzipFormat' with the default 'decompressAllMembers' @True@ you will
 -- have to do this, as the decompressor will look for any following members.
 -- With 'decompressAllMembers' @False@ the decompressor knows when the data
 -- ends and will produce 'DecompressStreamEnd' without you having to supply an
@@ -223,16 +235,20 @@ data DecompressError =
    | DictionaryRequired
 
      -- | If the stream requires a dictionary and you provide one with the
-     -- wrong 'DictionaryHash' then you will get this error.
+     -- wrong 'Stream.DictionaryHash' then you will get this error.
    | DictionaryMismatch
 
      -- | If the compressed data stream is corrupted in any way then you will
      -- get this error, for example if the input data just isn't a compressed
      -- zlib data stream. In particular if the data checksum turns out to be
      -- wrong then you will get all the decompressed data but this error at the
-     -- end, instead of the normal successful 'StreamEnd'.
+     -- end, instead of the normal successful 'Stream.StreamEnd'.
    | DataFormatError String
-  deriving (Eq, Typeable)
+  deriving
+  ( Eq
+  , Ord     -- ^ @since 0.7.0.0
+  , Generic -- ^ @since 0.7.0.0
+           )
 
 instance Show DecompressError where
   show TruncatedInput     = modprefix "premature end of compressed data stream"
@@ -283,7 +299,7 @@ foldDecompressStreamWithInput :: (S.ByteString -> a -> a)
                               -> L.ByteString
                               -> a
 foldDecompressStreamWithInput chunk end err = \s lbs ->
-    runST (fold s (L.toChunks lbs))
+    runST (fold s (toLimitedChunks lbs))
   where
     fold (DecompressInputRequired next) [] =
       next S.empty >>= \strm -> fold strm []
@@ -303,7 +319,8 @@ foldDecompressStreamWithInput chunk end err = \s lbs ->
 
 
 -- $incremental-compression
--- The pure 'compress' and 'decompress' functions are streaming in the sense
+-- The pure 'Codec.Compression.Zlib.Internal.compress' and
+-- 'Codec.Compression.Zlib.Internal.decompress' functions are streaming in the sense
 -- that they can produce output without demanding all input, however they need
 -- the input data stream as a lazy 'L.ByteString'. Having the input data
 -- stream as a lazy 'L.ByteString' often requires using lazy I\/O which is not
@@ -424,7 +441,7 @@ foldCompressStreamWithInput :: (S.ByteString -> a -> a)
                             -> L.ByteString
                             -> a
 foldCompressStreamWithInput chunk end = \s lbs ->
-    runST (fold s (L.toChunks lbs))
+    runST (fold s (toLimitedChunks lbs))
   where
     fold (CompressInputRequired next) [] =
       next S.empty >>= \strm -> fold strm []
@@ -452,10 +469,12 @@ compress   :: Stream.Format -> CompressParams -> L.ByteString -> L.ByteString
 -- | Incremental compression in the 'ST' monad. Using 'ST' makes it possible
 -- to write pure /lazy/ functions while making use of incremental compression.
 --
+-- Chunk size must fit into t'CUInt'.
 compressST :: Stream.Format -> CompressParams -> CompressStream (ST s)
 
 -- | Incremental compression in the 'IO' monad.
 --
+-- Chunk size must fit into t'CUInt'.
 compressIO :: Stream.Format -> CompressParams -> CompressStream IO
 
 compress   format params = foldCompressStreamWithInput
@@ -464,6 +483,7 @@ compress   format params = foldCompressStreamWithInput
 compressST format params = compressStreamST  format params
 compressIO format params = compressStreamIO  format params
 
+-- | Chunk size must fit into t'CUInt'.
 compressStream :: Stream.Format -> CompressParams -> S.ByteString
                -> Stream (CompressStream Stream)
 compressStream format (CompressParams compLevel method bits memLevel
@@ -476,8 +496,8 @@ compressStream format (CompressParams compLevel method bits memLevel
         if length == 0
           then fillBuffers 20   --gzip header is 20 bytes, others even smaller
           else do
-            Stream.pushInputBuffer inFPtr 0 length
-            fillBuffers initChunkSize
+            Stream.pushInputBuffer inFPtr 0 (int2cuint length)
+            fillBuffers (int2cuint_capped initChunkSize)
 
   where
     -- we flick between two states:
@@ -486,7 +506,7 @@ compressStream format (CompressParams compLevel method bits memLevel
     --   * where both buffers are non-empty
     --       - in which case we compress until a buffer is empty
 
-  fillBuffers :: Int -> Stream (CompressStream Stream)
+  fillBuffers :: CUInt -> Stream (CompressStream Stream)
   fillBuffers outChunkSize = do
 #ifdef DEBUG
     Stream.consistencyCheck
@@ -503,7 +523,7 @@ compressStream format (CompressParams compLevel method bits memLevel
     assert (inputBufferEmpty || outputBufferFull) $ return ()
 
     when outputBufferFull $ do
-      outFPtr <- Stream.unsafeLiftIO (S.mallocByteString outChunkSize)
+      outFPtr <- Stream.unsafeLiftIO (S.mallocByteString (cuint2int outChunkSize))
       Stream.pushOutputBuffer outFPtr 0 outChunkSize
 
     if inputBufferEmpty
@@ -511,7 +531,7 @@ compressStream format (CompressParams compLevel method bits memLevel
            if length == 0
              then drainBuffers True
              else do
-                Stream.pushInputBuffer inFPtr 0 length
+                Stream.pushInputBuffer inFPtr 0 (int2cuint length)
                 drainBuffers False
       else drainBuffers False
 
@@ -560,7 +580,10 @@ compressStream format (CompressParams compLevel method bits memLevel
   -- and if the format supports it (zlib and raw, not gzip).
   setDictionary :: Maybe S.ByteString -> Stream ()
   setDictionary (Just dict)
-    | Stream.formatSupportsDictionary format = do
+    | Stream.formatSupportsDictionary format = case int2cuint_safe (S.length dict) of
+      Nothing ->
+        fail "error when setting deflate dictionary, its length does not fit into CUInt"
+      Just{} -> do
         status <- Stream.deflateSetDictionary dict
         case status of
           Stream.Ok          -> return ()
@@ -580,10 +603,12 @@ decompress   :: Stream.Format -> DecompressParams -> L.ByteString -> L.ByteStrin
 -- | Incremental decompression in the 'ST' monad. Using 'ST' makes it possible
 -- to write pure /lazy/ functions while making use of incremental decompression.
 --
+-- Chunk size must fit into t'CUInt'.
 decompressST :: Stream.Format -> DecompressParams -> DecompressStream (ST s)
 
 -- | Incremental decompression in the 'IO' monad.
 --
+-- Chunk size must fit into t'CUInt'.
 decompressIO :: Stream.Format -> DecompressParams -> DecompressStream IO
 
 decompress   format params = foldDecompressStreamWithInput
@@ -592,7 +617,7 @@ decompress   format params = foldDecompressStreamWithInput
 decompressST format params = decompressStreamST  format params
 decompressIO format params = decompressStreamIO  format params
 
-
+-- | Chunk size must fit into t'CUInt'.
 decompressStream :: Stream.Format -> DecompressParams
                  -> Bool -> S.ByteString
                  -> Stream (DecompressStream Stream)
@@ -613,18 +638,17 @@ decompressStream format (DecompressParams bits initChunkSize mdict allMembers)
             -- special case to avoid demanding more input again
             -- always an error anyway
             when outputBufferFull $ do
-              let outChunkSize = 1
-              outFPtr <- Stream.unsafeLiftIO (S.mallocByteString outChunkSize)
-              Stream.pushOutputBuffer outFPtr 0 outChunkSize
+              outFPtr <- Stream.unsafeLiftIO (S.mallocByteString 1)
+              Stream.pushOutputBuffer outFPtr 0 1
             drainBuffers True
           else do
-            Stream.pushInputBuffer inFPtr 0 length
+            Stream.pushInputBuffer inFPtr 0 (int2cuint length)
             -- Normally we start with no output buffer (so counts as full) but
             -- if we're resuming then we'll usually still have output buffer
             -- space available
             assert (if not resume then outputBufferFull else True) $ return ()
             if outputBufferFull
-              then fillBuffers initChunkSize
+              then fillBuffers (int2cuint_capped initChunkSize)
               else drainBuffers False
 
   where
@@ -634,7 +658,7 @@ decompressStream format (DecompressParams bits initChunkSize mdict allMembers)
     --   * where both buffers are non-empty
     --       - in which case we compress until a buffer is empty
 
-  fillBuffers :: Int
+  fillBuffers :: CUInt
               -> Stream (DecompressStream Stream)
   fillBuffers outChunkSize = do
 #ifdef DEBUG
@@ -652,7 +676,7 @@ decompressStream format (DecompressParams bits initChunkSize mdict allMembers)
     assert (inputBufferEmpty || outputBufferFull) $ return ()
 
     when outputBufferFull $ do
-      outFPtr <- Stream.unsafeLiftIO (S.mallocByteString outChunkSize)
+      outFPtr <- Stream.unsafeLiftIO (S.mallocByteString (cuint2int outChunkSize))
       Stream.pushOutputBuffer outFPtr 0 outChunkSize
 
     if inputBufferEmpty
@@ -661,7 +685,7 @@ decompressStream format (DecompressParams bits initChunkSize mdict allMembers)
              if length == 0
                then drainBuffers True
                else do
-                 Stream.pushInputBuffer inFPtr 0 length
+                 Stream.pushInputBuffer inFPtr 0 (int2cuint length)
                  drainBuffers False
       else drainBuffers False
 
@@ -722,14 +746,16 @@ decompressStream format (DecompressParams bits initChunkSize mdict allMembers)
                 -> Stream (Maybe (DecompressStream Stream))
   setDictionary _adler Nothing =
     return $ Just (DecompressStreamError DictionaryRequired)
-  setDictionary _adler (Just dict) = do
-    status <- Stream.inflateSetDictionary dict
-    case status of
-      Stream.Ok -> return Nothing
-      Stream.Error Stream.DataError _   ->
-        return $ Just (DecompressStreamError DictionaryMismatch)
-      _ -> fail "error when setting inflate dictionary"
-
+  setDictionary _adler (Just dict) = case int2cuint_safe (S.length dict) of
+    Nothing ->
+      fail "error when setting inflate dictionary, its length does not fit into CUInt"
+    Just{} -> do
+      status <- Stream.inflateSetDictionary dict
+      case status of
+        Stream.Ok -> return Nothing
+        Stream.Error Stream.DataError _   ->
+          return $ Just (DecompressStreamError DictionaryMismatch)
+        _ -> fail "error when setting inflate dictionary"
 
 ------------------------------------------------------------------------------
 
@@ -743,6 +769,7 @@ runStreamIO :: Stream a -> Stream.State RealWorld -> IO (a, Stream.State RealWor
 runStreamST strm zstate = strictToLazyST (Unsafe.unsafeIOToST noDuplicate >> Stream.runStream strm zstate)
 runStreamIO strm zstate = stToIO (Stream.runStream strm zstate)
 
+-- | Chunk size must fit into t'CUInt'.
 compressStreamIO :: Stream.Format -> CompressParams -> CompressStream IO
 compressStreamIO format params =
     CompressInputRequired {
@@ -768,6 +795,7 @@ compressStreamIO format params =
 
     go CompressStreamEnd _ = CompressStreamEnd
 
+-- | Chunk size must fit into t'CUInt'.
 compressStreamST :: Stream.Format -> CompressParams -> CompressStream (ST s)
 compressStreamST format params =
     CompressInputRequired {
@@ -794,6 +822,7 @@ compressStreamST format params =
     go CompressStreamEnd _ = CompressStreamEnd
 
 
+-- | Chunk size must fit into t'CUInt'.
 decompressStreamIO :: Stream.Format -> DecompressParams -> DecompressStream IO
 decompressStreamIO format params =
       DecompressInputRequired $ \chunk -> do
@@ -873,6 +902,7 @@ decompressStreamIO format params =
         return (DecompressStreamError err)
 
 
+-- | Chunk size must fit into t'CUInt'.
 decompressStreamST :: Stream.Format -> DecompressParams -> DecompressStream (ST s)
 decompressStreamST format params =
       DecompressInputRequired $ \chunk -> do
@@ -925,8 +955,10 @@ decompressStreamST format params =
         if S.length chunk > 1
           then do
             -- have to handle the remaining data in this chunk
-            x <- runStreamST resume zstate
-            let (DecompressInputRequired next, zstate') = x
+            (x, zstate') <- runStreamST resume zstate
+            let next = case x of
+                  DecompressInputRequired n -> n
+                  _ -> error "checkHeaderSplit: unexpected result of runStreamST"
             (strm', zstate'') <- runStreamST (next (S.tail chunk)) zstate'
             go strm' zstate'' False
           else do
@@ -953,3 +985,24 @@ decompressStreamST format params =
     finaliseStreamError err zstate = do
         _ <- runStreamST Stream.finalise zstate
         return (DecompressStreamError err)
+
+-- | This one should not fail on 64-bit arch.
+cuint2int :: CUInt -> Int
+cuint2int n = fromMaybe (error $ "cuint2int: cannot cast " ++ show n) $ toIntegralSized n
+
+-- | This one could and will fail if chunks of ByteString are longer than 4G.
+int2cuint :: Int -> CUInt
+int2cuint n = fromMaybe (error $ "int2cuint: cannot cast " ++ show n) $ toIntegralSized n
+
+int2cuint_capped :: Int -> CUInt
+int2cuint_capped = fromMaybe maxBound . toIntegralSized . max 0
+
+int2cuint_safe :: Int -> Maybe CUInt
+int2cuint_safe = toIntegralSized
+
+toLimitedChunks :: L.ByteString -> [S.ByteString]
+toLimitedChunks L.Empty = []
+toLimitedChunks (L.Chunk x xs) = case int2cuint_safe (S.length x) of
+  Nothing -> let (y, z) = S.splitAt (cuint2int (maxBound :: CUInt)) x in
+    y : toLimitedChunks (L.Chunk z xs)
+  Just{} -> x : toLimitedChunks xs
