@@ -17,6 +17,7 @@ import Utils ()
 import Control.Monad
 import Control.Monad.ST.Lazy (ST)
 import Control.Exception
+import Data.Bits (finiteBitSize, shiftL)
 import qualified Data.ByteString.Char8 as BS.Char8
 import qualified Data.ByteString.Lazy as BL
 import qualified Data.ByteString      as BS
@@ -24,10 +25,6 @@ import qualified Data.ByteString      as BS
 import qualified Data.ByteString.Internal as BS
 #endif
 import System.IO
-#if !(MIN_VERSION_base(4,6,0))
-import Prelude hiding (catch)
-#endif
-
 
 main :: IO ()
 main = defaultMain $
@@ -43,17 +40,18 @@ main = defaultMain $
       testProperty "compress works with BSes with non-zero offset"   prop_compress_nonzero_bs_offset
     ],
     testGroup "unit tests" [
-      testProperty "simple gzip case"          test_simple_gzip,
-      testProperty "detect bad crc"            test_bad_crc,
-      testProperty "detect non-gzip"           test_non_gzip,
-      testProperty "detect custom dictionary"  test_custom_dict,
-      testProperty "dectect inflate with wrong dict"   test_wrong_dictionary,
-      testProperty "dectect inflate with right dict"   test_right_dictionary,
-      testProperty "handle trailing data"      test_trailing_data,
-      testProperty "multiple gzip members"     test_multiple_members,
-      testProperty "check small input chunks"  test_small_chunks,
-      testProperty "check empty input"         test_empty,
-      testProperty "check exception raised"    test_exception
+      testProperty "simple gzip case"               $ once test_simple_gzip,
+      testProperty "detect bad CRC"                 $ once test_bad_crc,
+      testProperty "detect non-gzip"                $ once test_non_gzip,
+      testProperty "detect custom dictionary"       $ once test_custom_dict,
+      testProperty "detect inflate with wrong dict" $ once test_wrong_dictionary,
+      testProperty "detect inflate with right dict" $ once test_right_dictionary,
+      testProperty "handle trailing data"           $ once test_trailing_data,
+      testProperty "multiple gzip members"          $ once test_multiple_members,
+      testProperty "check small input chunks"       $ once test_small_chunks,
+      testProperty "check empty input"              $ once test_empty,
+      testProperty "check exception raised"         $ once test_exception,
+      testProperty "check compress large chunk"     $ once test_compress_large_chunk
     ]
   ]
 
@@ -285,6 +283,18 @@ test_exception = ioProperty $ do
     Left err -> assertDataFormatError "incorrect data check" err
     Right{} -> counterexample "expected exception" False
 
+test_compress_large_chunk :: Property
+test_compress_large_chunk =
+  counterexample
+    ("Expected " ++ show len ++ " zeros but got different result, please investigate manually")
+    (property test)
+  where
+    test = GZip.decompress (GZip.compress (BL.replicate len 0)) == BL.replicate len 0
+
+    len = case finiteBitSize (0 :: Int) of
+      64 -> (1 `shiftL` 32) + 1
+      _ -> 0 -- ignore it
+
 toStrict :: BL.ByteString -> BS.ByteString
 #if MIN_VERSION_bytestring(0,10,0)
 toStrict = BL.toStrict
